/*
 * cloudGeometry
 *
 * Evan Herbst
 * 6 / 23 / 10
 */

#include <cassert>
#include <cfloat>
#include "rgbd_util/eigen/QR" //eigensolving
#include "rgbd_util/eigen/Eigenvalues"
#include "pcl_rgbd/pointTypes.h"

namespace rgbd
{

template <typename PointT>
rgbd::eigen::Vector3f centroid(const pcl::PointCloud<PointT>& cloud)
{
	rgbd::eigen::Vector3f mean(0, 0, 0);
	for(unsigned int i = 0; i < cloud.points.size(); i++) mean += rgbd::ptX2eigen<rgbd::eigen::Vector3f>(cloud.points[i]);
	return mean / cloud.points.size();
}

template <typename PointT>
rgbd::eigen::Vector3f covariance(const pcl::PointCloud<PointT>& cloud)
{
	const rgbd::eigen::Vector3f mean = centroid(cloud);

	rgbd::eigen::Matrix3f cov; //covariance
	cov.fill(0);

	/*
	 * compute below-diagonal entries
	 */
	for(unsigned int k = 0; k < cloud.points.size(); k++)
	{
		const float x[3] = {cloud.points[k].x, cloud.points[k].y, cloud.points[k].z};
		for(unsigned int i = 0; i < 3; i++)
			for(unsigned int j = 0; j <= i; j++)
				cov(i, j) += x[i] * x[j];
	}
	for(unsigned int i = 0; i < 3; i++)
		for(unsigned int j = 0; j <= i; j++)
			cov(i, j) = cov(i, j) / cloud.points.size() - mean[i] * mean[j];

	for(unsigned int i = 0; i < 3; i++) //copy to above-diagonal entries
		for(unsigned int j = i + 1; j < 3; j++)
			cov(i, j) = cov(j, i);

	return cov;
}

/*
 * bounding box: min (x y z), max (x y z)
 */
template <typename PointT>
std::tuple<rgbd::eigen::Vector3f, rgbd::eigen::Vector3f> bbox(const pcl::PointCloud<PointT>& cloud)
{
	assert(!cloud.points.empty());
	std::tuple<rgbd::eigen::Vector3f, rgbd::eigen::Vector3f> result;
	rgbd::eigen::Vector3f& mins = std::get<0>(result), &maxes = std::get<1>(result);
	mins.fill(FLT_MAX);
	maxes.fill(-FLT_MAX);
	for(auto i = cloud.points.begin(); i != cloud.points.end(); i++)
	{
		if((*i).x < mins[0]) mins[0] = (*i).x;
		if((*i).y < mins[1]) mins[1] = (*i).y;
		if((*i).z < mins[2]) mins[2] = (*i).z;
		if((*i).x < maxes[0]) maxes[0] = (*i).x;
		if((*i).y < maxes[1]) maxes[1] = (*i).y;
		if((*i).z < maxes[2]) maxes[2] = (*i).z;
	}
	return result;
}

/*
 * return all eigenvectors of the covariance mtx, one per row
 */
template <typename PointT>
rgbd::eigen::Matrix3f eigenvectors(const pcl::PointCloud<PointT>& cloud)
{
	const rgbd::eigen::Matrix3f cov = covariance(cloud);
	rgbd::eigen::SelfAdjointEigenSolver<rgbd::eigen::Matrix3f> eigsolver(cov);
	const rgbd::eigen::Matrix3f eigvecs = eigsolver.eigenvectors(); //one per column
	return eigvecs.transpose();
}

} //namespace
