/*
 * cloudTofroPLY: write PointClouds to and read them from PLY files
 *
 * Evan Herbst
 * 6 / 16 / 10
 */

#include <sensor_msgs/point_cloud_conversion.h>
#include <pcl/ros/conversions.h>
#include "pcl_rgbd/cloudUtils.h"

namespace rgbd
{

/**
 * throw on any error
 */
template <typename PointT>
void write_ply_file(const pcl::PointCloud<PointT>& cloud, const fs::path& filepath)
{
	sensor_msgs::PointCloud2 msg;
	pcl::toROSMsg(cloud, msg);
	write_ply_file(msg, filepath);
}

/*
 * edit the data of cloud but not the header
 *
 * throw on any format error
 */
template <typename PointT>
void read_ply_file(pcl::PointCloud<PointT>& cloud, const fs::path& filepath)
{
	sensor_msgs::PointCloud2 msg;
	read_ply_file(msg, filepath);
	pcl::fromROSMsg(msg, cloud);
}

/*
 * deal with format conversion of PLYs written by write_ply_file()
 *
 * PointT should be rgbd::pt or rgbd::surfelPt or similar
 */
template <typename PointT>
typename pcl::PointCloud<PointT>::Ptr readRGBDPLY(const fs::path& filepath)
{
	sensor_msgs::PointCloud2::Ptr cloud(new sensor_msgs::PointCloud2);
	rgbd::read_ply_file(*cloud, filepath.string());
	sensor_msgs::PointCloud2::Ptr cloud2(new sensor_msgs::PointCloud2);
	if(sensor_msgs::getPointCloud2FieldIndex(*cloud, "rgb") == -1) rgbd::packColorChannels("diffuse_red", "diffuse_green", "diffuse_blue", *cloud, *cloud2);
	else cloud2.swap(cloud);

	const int imgXFieldNum = sensor_msgs::getPointCloud2FieldIndex(*cloud2, "imgX");
	if(imgXFieldNum >= 0 && cloud2->fields[imgXFieldNum].datatype == sensor_msgs::PointField::FLOAT32)
	{
		rgbd::convertFieldType<float, uint32_t>("imgX", *cloud2, *cloud);
		rgbd::convertFieldType<float, uint32_t>("imgY", *cloud, *cloud2);
	}

	typename pcl::PointCloud<PointT>::Ptr surfelCloud(new pcl::PointCloud<PointT>);
	pcl::fromROSMsg(*cloud2, *surfelCloud);
	return surfelCloud;
}

}
