/*
 * normalsIO: I/O for point-cloud normals
 *
 * Evan Herbst
 * 12 / 20 / 10
 */

#include <cassert>
#include <fstream>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/gzip.hpp>
#include "rgbd_util/ioUtils.h"
#include "pcl_rgbd/normalsIO.h"

namespace rgbd
{
namespace io = boost::iostreams;

template <typename PointT>
void writeNormals(const pcl::PointCloud<PointT>& cloud, const fs::path& filepath)
{
	std::ofstream outfile(filepath.string().c_str(), std::ofstream::binary);
	assert(outfile);
	io::filtering_ostream outfilt;
	outfilt.push(io::gzip_compressor());
	outfilt.push(outfile);
	writeBinary(outfilt, (uint64_t)cloud.width);
	writeBinary(outfilt, (uint64_t)cloud.height);
	for(unsigned int i = 0; i < cloud.points.size(); i++)
		for(unsigned int j = 0; j < 3; j++)
			writeBinary(outfilt, cloud.points[i].normal[j]);
}

/*
 * cloud will be resized if nec
 *
 * only the normals fields of cloud will be set
 */
template <typename PointT>
void readNormals(const fs::path& filepath, pcl::PointCloud<PointT>& cloud)
{
	std::ifstream infile(filepath.string().c_str(), std::ifstream::binary);
	assert(infile);
	io::filtering_istream infilt;
	infilt.push(io::gzip_decompressor());
	infilt.push(infile);
	uint64_t w, h;
	readBinary(infilt, w);
	readBinary(infilt, h);
	cloud.points.resize(w * h);
	cloud.width = w;
	cloud.height = h;
	for(unsigned int i = 0; i < cloud.points.size(); i++)
		for(unsigned int j = 0; j < 3; j++)
			readBinary(infilt, cloud.points[i].normal[j]);
}

} //namespace
