
/*
 * to compile this code, you need to have NEIGHBORS_OFFSET define set
 * defining MAX_Z is optional
 * warning: the curvature calculation is probably unreliable
 */


// source for eigenvector computation:
// http://barnesc.blogspot.com/2007/02/eigenvectors-of-3x3-symmetric-matrix.html


/* Eigen decomposition code for symmetric 3x3 matrices, copied from the public
   domain Java Matrix library JAMA. */

#ifdef MAX
#undef MAX
#endif

#define MAX(a, b) ((a)>(b)?(a):(b))

#define n 3

__private float hypot2(__private float x, __private float y) {
	return sqrt(x*x+y*y);
}

// Symmetric Householder reduction to tridiagonal form.
void tred2(__private float V[n][n], __private float d[n], __private float e[n]) {

//  This is derived from the Algol procedures tred2 by
//  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
//  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
//  Fortran subroutine in EISPACK.

  for (int j = 0; j < n; j++) {
    d[j] = V[n-1][j];
  }

  // Householder reduction to tridiagonal form.

  for (int i = n-1; i > 0; i--) {

    // Scale to avoid under/overflow.

    float scale = 0.0;
    float h = 0.0;
    for (int k = 0; k < i; k++) {
      scale = scale + fabs(d[k]);
    }
    if (scale == 0.0) {
      e[i] = d[i-1];
      for (int j = 0; j < i; j++) {
        d[j] = V[i-1][j];
        V[i][j] = 0.0;
        V[j][i] = 0.0;
      }
    } else {

      // Generate Householder vector.

      for (int k = 0; k < i; k++) {
        d[k] /= scale;
        h += d[k] * d[k];
      }
      float f = d[i-1];
      float g = sqrt(h);
      if (f > 0) {
        g = -g;
      }
      e[i] = scale * g;
      h = h - f * g;
      d[i-1] = f - g;
      for (int j = 0; j < i; j++) {
        e[j] = 0.0;
      }

      // Apply similarity transformation to remaining columns.

      for (int j = 0; j < i; j++) {
        f = d[j];
        V[j][i] = f;
        g = e[j] + V[j][j] * f;
        for (int k = j+1; k <= i-1; k++) {
          g += V[k][j] * d[k];
          e[k] += V[k][j] * f;
        }
        e[j] = g;
      }
      f = 0.0;
      for (int j = 0; j < i; j++) {
        e[j] /= h;
        f += e[j] * d[j];
      }
      float hh = f / (h + h);
      for (int j = 0; j < i; j++) {
        e[j] -= hh * d[j];
      }
      for (int j = 0; j < i; j++) {
        f = d[j];
        g = e[j];
        for (int k = j; k <= i-1; k++) {
          V[k][j] -= (f * e[k] + g * d[k]);
        }
        d[j] = V[i-1][j];
        V[i][j] = 0.0;
      }
    }
    d[i] = h;
  }

  // Accumulate transformations.

  for (int i = 0; i < n-1; i++) {
    V[n-1][i] = V[i][i];
    V[i][i] = 1.0;
    float h = d[i+1];
    if (h != 0.0) {
      for (int k = 0; k <= i; k++) {
        d[k] = V[k][i+1] / h;
      }
      for (int j = 0; j <= i; j++) {
        float g = 0.0;
        for (int k = 0; k <= i; k++) {
          g += V[k][i+1] * V[k][j];
        }
        for (int k = 0; k <= i; k++) {
          V[k][j] -= g * d[k];
        }
      }
    }
    for (int k = 0; k <= i; k++) {
      V[k][i+1] = 0.0;
    }
  }
  for (int j = 0; j < n; j++) {
    d[j] = V[n-1][j];
    V[n-1][j] = 0.0;
  }
  V[n-1][n-1] = 1.0;
  e[0] = 0.0;
} 

// Symmetric tridiagonal QL algorithm.

void tql2(__private float V[n][n], __private float d[n], __private float e[n]) {

//  This is derived from the Algol procedures tql2, by
//  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
//  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
//  Fortran subroutine in EISPACK.

  for (int i = 1; i < n; i++) {
    e[i-1] = e[i];
  }
  e[n-1] = 0.0;

  float f = 0.0;
  float tst1 = 0.0;
  float eps = pow((float)2.0,(float)-52.0);
  for (int l = 0; l < n; l++) {

    // Find small subdiagonal element

    tst1 = MAX(tst1,fabs(d[l]) + fabs(e[l]));
    int m = l;
    while (m < n) {
      if (fabs(e[m]) <= eps*tst1) {
        break;
      }
      m++;
    }

    // If m == l, d[l] is an eigenvalue,
    // otherwise, iterate.

    if (m > l) {
      int iter = 0;
      do {
        iter = iter + 1;  // (Could check iteration count here.)

        // Compute implicit shift

        float g = d[l];
        float p = (d[l+1] - g) / (2.0 * e[l]);
        float r = hypot2(p,1.0);
        if (p < 0) {
          r = -r;
        }
        d[l] = e[l] / (p + r);
        d[l+1] = e[l] * (p + r);
        float dl1 = d[l+1];
        float h = g - d[l];
        for (int i = l+2; i < n; i++) {
          d[i] -= h;
        }
        f = f + h;

        // Implicit QL transformation.

        p = d[m];
        float c = 1.0;
        float c2 = c;
        float c3 = c;
        float el1 = e[l+1];
        float s = 0.0;
        float s2 = 0.0;
        for (int i = m-1; i >= l; i--) {
          c3 = c2;
          c2 = c;
          s2 = s;
          g = c * e[i];
          h = c * p;
          r = hypot2(p,e[i]);
          e[i+1] = s * r;
          s = e[i] / r;
          c = p / r;
          p = c * d[i] - s * g;
          d[i+1] = h + s * (c * g + s * d[i]);

          // Accumulate transformation.

          for (int k = 0; k < n; k++) {
            h = V[k][i+1];
            V[k][i+1] = s * V[k][i] + c * h;
            V[k][i] = c * V[k][i] - s * h;
          }
        }
        p = -s * s2 * c3 * el1 * e[l] / dl1;
        e[l] = s * p;
        d[l] = c * p;

        // Check for convergence.

      } while (fabs(e[l]) > eps*tst1);
    }
    d[l] = d[l] + f;
    e[l] = 0.0;
  }
  
  // Sort eigenvalues and corresponding vectors.

  for (int i = 0; i < n-1; i++) {
    int k = i;
    float p = d[i];
    for (int j = i+1; j < n; j++) {
      if (d[j] < p) {
        k = j;
        p = d[j];
      }
    }
    if (k != i) {
      d[k] = d[i];
      d[i] = p;
      for (int j = 0; j < n; j++) {
        p = V[j][i];
        V[j][i] = V[j][k];
        V[j][k] = p;
      }
    }
  }
}


//
// all code following by Polly Powledge, Intel Labs Seattle, 2010
//

#define X_ELEMENT(v4)	v4.x
#define Y_ELEMENT(v4)	v4.y
#define Z_ELEMENT(v4)	v4.z
#define W_ELEMENT(v4)	v4.w

#define MIN_NN			4
#define FOCAL_DIST		570.342

#if USE_DEBUG_BUFFER
__kernel void compute_normals( __read_only image2d_t point_data, float viewpoint_x, float viewpoint_y, float viewpoint_z, __global float4 *output_data, __global float *debug_data )
#else
__kernel void compute_normals( __read_only image2d_t point_data, float viewpoint_x, float viewpoint_y, float viewpoint_z, __global float4 *output_data )
#endif
{

	const int num_threads_in_wg = get_local_size(1);
	const int wg_id = get_group_id(0);
	const int target_x = get_global_id(0);
	const int target_y = get_global_id(1);

#if 0  
int offset = (target_y * 640) + target_x;
float4 stuff = { target_x, target_y, offset, 555 };
*(output_data + offset) = stuff;
return;
#endif

	// this code supports 640x480 images only
	if ( target_y > 479 ) return;
#if 0 
if ( target_x != 492 || target_y != 10 ) return;
//if ( target_x != 593 || target_y != 9 ) return;
//if ( target_x != 601 || target_y != 9 ) return;
int kkk = 0;
#endif

	////////////////////////////////////////////////////////////////////////////////////////
	// step 0: read target pixel's z value. if it's less than or equal to 0, exit
	////////////////////////////////////////////////////////////////////////////////////////
	sampler_t sampler = CLK_NORMALIZED_COORDS_FALSE|CLK_ADDRESS_NONE|CLK_FILTER_NEAREST;
	float4 target_pixel = read_imagef(point_data,sampler,(int2)(target_x,target_y));
	//float4 target_pixel = read_imagef(point_data,sampler,(int2)(target_y,target_x));
#if 0 
float4 su = { target_x, target_y, target_pixel.x, 333 };
*(output_data + kkk) = su;
kkk++;
#endif

	if ( target_pixel.x <= 0.0f ) 
	{
		float4 missing_data = { 9999, 9999, target_x, target_y };
		*(output_data + (target_y * 640) + target_x) = missing_data;
		return;
	}
	float new_target_x = (( target_x - (640/2)) * target_pixel.x /  FOCAL_DIST);
	float new_target_y = (( target_y - (480/2)) * target_pixel.x /  FOCAL_DIST);

	////////////////////////////////////////////////////////////////////////////////////////
	// step 1: got a valid point -- now compute its centroid
	////////////////////////////////////////////////////////////////////////////////////////
	int startx = target_x - NEIGHBORS_OFFSET;
	int starty = target_y - NEIGHBORS_OFFSET;
	int endx = target_x + NEIGHBORS_OFFSET;
	int endy = target_y + NEIGHBORS_OFFSET;

	startx = startx<0?0:startx;
	starty = starty<0?0:starty;
	endx = endx>639?639:endx;
	endy = endy>479?479:endy;

#if 0 
float4 dd = { target_x, startx, endx, 1212 };
*(output_data + kkk) = dd;
kkk++;
float4 ddd = { target_y, starty, endy, 2121 };
*(output_data + kkk) = ddd;
kkk++;
#endif

	__private float4 centroid_data = { 0.0f, 0.0f, 0.0f, 0.0f };
	__private float4 pixel;
	__private int num_valid_pts = 0;

	for ( int y = starty ; y <= endy ; y++ )
	{
		for ( int x = startx ; x <= endx ; x++ )
		{
			pixel = read_imagef(point_data,sampler,(int2)(x,y));
			//pixel = read_imagef(point_data,sampler,(int2)(y,x));
#ifdef MAX_Z
			if ( fabs( pixel.x - target_pixel.x ) > MAX_Z ) 
			{
#if 0 
float4 vd = { pixel.x, target_pixel.x ,MAX_Z, 9999.1 };
*(output_data + kkk) = vd;
kkk++;
#endif
				continue; 
			}
#endif
			if ( pixel.x <= 0.0f ) continue;

			float newx = (( x - (640/2)) * pixel.x /  FOCAL_DIST);
			float newy = (( y - (480/2)) * pixel.x /  FOCAL_DIST);
			X_ELEMENT(centroid_data) += newx;
			Y_ELEMENT(centroid_data) += newy;
			Z_ELEMENT(centroid_data) += pixel.x;
#if 0 
float4 d = { newx, newy ,pixel.x, 5454.1 };
*(output_data + kkk) = d;
kkk++;
#endif
			num_valid_pts++;
		}
	}

#if 0 
float4 num = { num_valid_pts, num_valid_pts, num_valid_pts, 7777.1 };
*(output_data + kkk) = num;
kkk++;
//return;
#endif

	if (  num_valid_pts < MIN_NN )
	{
		float4 not_enuf_nn = { 8888, 8888, target_x, target_y };
		*(output_data + ((target_x * 640) + target_y)) = not_enuf_nn;
		return;
	}

#if 0 
float4 cd = { centroid_data.x, centroid_data.y, centroid_data.z, 6666.1 };
*(output_data + kkk) = cd;
kkk++;
#endif

	X_ELEMENT(centroid_data) /= num_valid_pts;
	Y_ELEMENT(centroid_data) /= num_valid_pts;
	Z_ELEMENT(centroid_data) /= num_valid_pts;

#define DUMP_CENTROID 0	
#if DUMP_CENTROID
	float4 d = { centroid_data.x, centroid_data.y, centroid_data.z, 6666.2 };
	*(output_data + kkk) = d;
	return;
#endif

	////////////////////////////////////////////////////////////////////////////////////////
	// step 2: compute the covariance matrix
	////////////////////////////////////////////////////////////////////////////////////////
	__private float cov_matrix[3][3];
	cov_matrix[0][0] = 0.0f;
	cov_matrix[0][1] = 0.0f;
	cov_matrix[0][2] = 0.0f;
	cov_matrix[1][0] = 0.0f;
	cov_matrix[1][1] = 0.0f;
	cov_matrix[1][2] = 0.0f;
	cov_matrix[2][0] = 0.0f;
	cov_matrix[2][1] = 0.0f;
	cov_matrix[2][2] = 0.0f;

	float4 nn_point;
	for ( int y = starty ; y <= endy ; y++ )
	{
		for ( int x = startx ; x <= endx ; x++ )
		{
			//if ( x == target_x && y == target_y ) continue; 
			pixel = read_imagef(point_data,sampler,(int2)(x,y));
			//pixel = read_imagef(point_data,sampler,(int2)(y,x));
#ifdef MAX_Z
			if ( fabs( pixel.x - target_pixel.x ) > MAX_Z ) continue; 
#endif
			if ( pixel.x <= 0.0f ) continue;
			X_ELEMENT(nn_point) = (( x - (640/2)) * pixel.x /  FOCAL_DIST);
			Y_ELEMENT(nn_point) = (( y - (480/2)) * pixel.x /  FOCAL_DIST);
			Z_ELEMENT(nn_point) = pixel.x;
			cov_matrix[0][0] += ( (X_ELEMENT(nn_point) - X_ELEMENT(centroid_data)) * (X_ELEMENT(nn_point) - X_ELEMENT(centroid_data)) );
			cov_matrix[0][1] += ( (X_ELEMENT(nn_point) - X_ELEMENT(centroid_data)) * (Y_ELEMENT(nn_point) - Y_ELEMENT(centroid_data)) );
			cov_matrix[0][2] += ( (X_ELEMENT(nn_point) - X_ELEMENT(centroid_data)) * (Z_ELEMENT(nn_point) - Z_ELEMENT(centroid_data)) );

			cov_matrix[1][0] += ( (Y_ELEMENT(nn_point) - Y_ELEMENT(centroid_data)) * (X_ELEMENT(nn_point) - X_ELEMENT(centroid_data)) );
			cov_matrix[1][1] += ( (Y_ELEMENT(nn_point) - Y_ELEMENT(centroid_data)) * (Y_ELEMENT(nn_point) - Y_ELEMENT(centroid_data)) );
			cov_matrix[1][2] += ( (Y_ELEMENT(nn_point) - Y_ELEMENT(centroid_data)) * (Z_ELEMENT(nn_point) - Z_ELEMENT(centroid_data)) );

			cov_matrix[2][0] += ( (Z_ELEMENT(nn_point) - Z_ELEMENT(centroid_data)) * (X_ELEMENT(nn_point) - X_ELEMENT(centroid_data)) );
			cov_matrix[2][1] += ( (Z_ELEMENT(nn_point) - Z_ELEMENT(centroid_data)) * (Y_ELEMENT(nn_point) - Y_ELEMENT(centroid_data)) );
			cov_matrix[2][2] += ( (Z_ELEMENT(nn_point) - Z_ELEMENT(centroid_data)) * (Z_ELEMENT(nn_point) - Z_ELEMENT(centroid_data)) );

#if 0 
float4 cm = { X_ELEMENT(nn_point), Y_ELEMENT(nn_point), Z_ELEMENT(nn_point), 33 };
*(output_data + kkk ) = cm;
kkk++;
#endif

		}
	}

#define DUMP_COV_MATRIX 0	
#if DUMP_COV_MATRIX
	float4 cm1 = { cov_matrix[0][0], cov_matrix[0][1], cov_matrix[0][2], 0.0 };
	*(output_data + kkk ) = cm1;
	kkk++;
	float4 cm2 = { cov_matrix[1][0], cov_matrix[1][1], cov_matrix[1][2], 0.0 };
	*(output_data + kkk ) = cm2;
	kkk++;
	float4 cm3 = { cov_matrix[2][0], cov_matrix[2][1], cov_matrix[2][2], 0.0 };
	*(output_data + kkk ) = cm3;
	kkk++;
	return;
#endif


	////////////////////////////////////////////////////////////////////////////////////////
	// step 3: compute the eigen pair
	////////////////////////////////////////////////////////////////////////////////////////

	// d_eg in is the input symmetric 3x3 matrix
	// d_eg out is where the computed eigenvectors are stored. Each vector is stored as
	// a column. values are normed to absolute value 1
	__private float d_eg[3][3];
	// d_ev out are the computed corresponding eigenvalues. smallest ev is in offset 0. 
	__private float d_ev[3];

#if 1
	d_eg[0][0] = cov_matrix[0][0];
	d_eg[0][1] = cov_matrix[0][1];
	d_eg[0][2] = cov_matrix[0][2];
	d_eg[1][0] = cov_matrix[1][0];
	d_eg[1][1] = cov_matrix[1][1];
	d_eg[1][2] = cov_matrix[1][2];
	d_eg[2][0] = cov_matrix[2][0];
	d_eg[2][1] = cov_matrix[2][1];
	d_eg[2][2] = cov_matrix[2][2];
#else
	// test data
	d_eg[0][0] = 0.0f;
	d_eg[0][1] = 1.0f;
	d_eg[0][2] = -1.0f;
	d_eg[1][0] = 1.0f;
	d_eg[1][1] = 1.0f;
	d_eg[1][2] = 0.0f;
	d_eg[2][0] = -1.0f;
	d_eg[2][1] = 0.0f;
	d_eg[2][2] = 1.0f;
	// ev's should be 2, 1, -1
#endif

	__private float e[3];
	tred2(d_eg, d_ev, e);
	tql2(d_eg, d_ev, e);

#define DUMP_EV 0	
#if DUMP_EV
	float4 d = { d_eg[0][0], d_eg[1][0], d_eg[2][0], d_ev[0] };
	*(output_data ) = d;
	d.x = d_eg[0][1];
	d.y = d_eg[1][1];
	d.z = d_eg[2][1];
	d.w = d_ev[1];
	*(output_data + 1 ) = d;
	d.x = d_eg[0][2];
	d.y = d_eg[1][2];
	d.z = d_eg[2][2];
	d.w = d_ev[2];
	*(output_data + 2 ) = d;
	return;
#endif

	////////////////////////////////////////////////////////////////////////////////////////
	// step 4: normalize the surface normal, and compute the curvature
	////////////////////////////////////////////////////////////////////////////////////////

	float norm = sqrt( (d_eg[0][0] * d_eg[0][0]) + (d_eg[1][0] * d_eg[1][0]) + (d_eg[2][0] * d_eg[2][0]) );
	float4 plane_parameters = { d_eg[0][0]/norm, d_eg[1][0]/norm, d_eg[2][0]/norm };
	// Hessian form (D = nc . p_plane (centroid here) + p)
	W_ELEMENT(plane_parameters) = -1 * ( (X_ELEMENT(plane_parameters) * X_ELEMENT(centroid_data)) + (Y_ELEMENT(plane_parameters) * Y_ELEMENT(centroid_data)) + (Z_ELEMENT(plane_parameters) * Z_ELEMENT(centroid_data)) );
	float curvature = fabs( d_ev[0] / ( d_ev[0] + d_ev[1] + d_ev[2] ) );

#define DUMP_PP	0	
#if DUMP_PP
	float4 pp = { X_ELEMENT(plane_parameters), Y_ELEMENT(plane_parameters), Z_ELEMENT(plane_parameters), W_ELEMENT(plane_parameters) };
	*(output_data + kkk ) = pp;
	kkk++;
	//return;
#endif
	
	////////////////////////////////////////////////////////////////////////////////////////
	// step 5: flip the normal towards the viewpoint
	////////////////////////////////////////////////////////////////////////////////////////
	float4 adj_vp;

	X_ELEMENT(adj_vp) = viewpoint_x - new_target_x;
	Y_ELEMENT(adj_vp) = viewpoint_y - new_target_y;
	Z_ELEMENT(adj_vp) = viewpoint_z - target_pixel.x;

	float cos_theta = ( X_ELEMENT(adj_vp) * X_ELEMENT(plane_parameters) ) + ( Y_ELEMENT(adj_vp) * Y_ELEMENT(plane_parameters) ) + (Z_ELEMENT(adj_vp) * Z_ELEMENT(plane_parameters) );
#define DUMP_CT	0	
#if DUMP_CT
	float4 pp = { X_ELEMENT(plane_parameters), Y_ELEMENT(plane_parameters), Z_ELEMENT(plane_parameters), W_ELEMENT(plane_parameters) };
	*(output_data  ) = pp;
	float4 ct = { X_ELEMENT(adj_vp), Y_ELEMENT(adj_vp), Z_ELEMENT(adj_vp), cos_theta };
	*(output_data + 1 ) = ct;
	float4 ct2 = { 8888, cos_theta, cos_theta, cos_theta };
	*(output_data + 2 ) = ct2;
	float4 nt = { 9999, new_target_x, new_target_y, target_pixel[0] };
	*(output_data + 3 ) = nt;
	float4 vp = { 1111, viewpoint_x, viewpoint_y, viewpoint_z };
	*(output_data + 4 ) = vp;
	kkk++;
	return;
#endif

	////////////////////////////////////////////////////////////////////////////////////////
	// warning: the curvature calculation needs some fixin' ... don't trust it
	////////////////////////////////////////////////////////////////////////////////////////
	if ( cos_theta < 0 )
	{
		X_ELEMENT(plane_parameters) *= -1;
		Y_ELEMENT(plane_parameters) *= -1;
		Z_ELEMENT(plane_parameters) *= -1;
		// Hessian form (D = nc . p_plane (centroid here) + p)
		W_ELEMENT(plane_parameters) = -1 * ( (X_ELEMENT(plane_parameters) * new_target_x) + (Y_ELEMENT(plane_parameters) * new_target_y) + (Z_ELEMENT(plane_parameters) * target_pixel.x) );
	}
		
	*(output_data + (target_y * 640) + target_x) = plane_parameters;
//*(output_data ) = plane_parameters;

}


