/*
 * icpPlanarPatch: ICP planar-patch-to-planar-patch error function
 *
 * Evan Herbst
 * 7 / 19 / 10
 */

#include <cmath> //acos()
#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>

#define POINT_VECTOR_STUFF \
	BOOST_STATIC_ASSERT((boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector3f> >::value || boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector4f, rgbd::eigen::aligned_allocator<rgbd::eigen::Vector4f> > >::value));\
	typedef typename PointVectorT::value_type PointT;

namespace icp
{
namespace planarPatch
{

template <typename PointVectorT>
void getSourceErrorVector(const PointVectorT& sourcePoints, const PointVectorT& sourceNormals, const PointVectorT& targetPoints, const PointVectorT& targetNormals, const std::vector<float>& pointWeights,
									rgbd::eigen::Affine3f const& transform, const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(2 * point_count);

	for (unsigned int i = 0; i < point_count; ++i)
		if(correspondence_indices[i] >= 0)
		{
			/*
			 * angle between normals
			 */
			const PointT xformedNormal = transform * sourceNormals[i];
			const float ang = acos(targetNormals[correspondence_indices[i]].dot(xformedNormal));
			result.push_back(pointWeights[i] * ang);

			/*
			 * dist of transformed pt from target plane
			 */
			const PointT xformedPt = transform * sourcePoints[i];
			const float pt2planeDist = (xformedPt - targetPoints[correspondence_indices[i]]).dot(targetNormals[correspondence_indices[i]]);
			result.push_back(pointWeights[i] * pt2planeDist * 1.67/* Dieter recommends that 10 cm distance error be equivalent to 10 degrees angular error */);
		}
}
template <typename PointVectorT>
void getTargetErrorVector(const PointVectorT& targetPoints, const PointVectorT& targetNormals, const std::vector<float>& pointWeights,
									const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();

	result.clear();
	result.reserve(2 * point_count);
	for (unsigned int i = 0; i < point_count; ++i)
		if(correspondence_indices[i] >= 0)
		{
			result.push_back(0);
			result.push_back(0);
		}
}

} //namespace
} //namespace

#undef POINT_VECTOR_STUFF
