/*
 * icpPoint2Line
 *
 * Provides source and target error vectors for ICP
 *
 * Evan Herbst
 * 7 / 10 / 10
 */

#include <cmath> //sqrt()
#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>

#define POINT_VECTOR_STUFF \
	BOOST_STATIC_ASSERT((boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector3f> >::value || boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector4f, rgbd::eigen::aligned_allocator<rgbd::eigen::Vector4f> > >::value));\
	typedef typename PointVectorT::value_type PointT;

namespace icp
{
namespace point2line
{

template <typename PointVectorT>
void getSourceErrorVector(const PointVectorT& sourcePoints, const PointVectorT& targetNormals1, const NormalVectorT& targetNormals2,
	const std::vector<float>& pointWeights, rgbd::eigen::Affine3f const& transform, const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(2*point_count);

	// point-to-line
	for (unsigned int i = 0; i < point_count; ++i) {
		if (correspondence_indices[i] >= 0) {
			PointT transformed_source = transform * sourcePoints[i];
			float dot_product1 = targetNormals1[correspondence_indices[i]].dot(transformed_source);
			float dot_product2 = targetNormals2[correspondence_indices[i]].dot(transformed_source);
			result.push_back(pointWeights[i]*dot_product1);
			result.push_back(pointWeights[i]*dot_product2);
		}
	}
}

template <typename PointVectorT>
void getTargetErrorVector(const PointVectorT& targetPoints, const PointVectorT& targetNormals1, const NormalVectorT& targetNormals2,
	const std::vector<float>& pointWeights, const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(2*point_count);

	// point-to-line
	for (unsigned int i = 0; i < point_count; ++i) {
		if (correspondence_indices[i] >= 0) {
			PointT target = targetPoints[correspondence_indices[i]];
			float dot_product1 = targetNormals1[correspondence_indices[i]].dot(target);
			float dot_product2 = targetNormals2[correspondence_indices[i]].dot(target);
			result.push_back(pointWeights[i] * dot_product1);
			result.push_back(pointWeights[i] * dot_product2);
		}
	}
}

} // ns
} // ns

#undef POINT_VECTOR_STUFF
