/*
 * icpPoint2Point
 *
 * Provides source and target error vectors for ICP
 *
 * Evan Herbst
 * 7 / 10 / 10
 */

#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>

#define POINT_VECTOR_STUFF \
	BOOST_STATIC_ASSERT((boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector3f> >::value || boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector4f, rgbd::eigen::aligned_allocator<rgbd::eigen::Vector4f> > >::value));\
	typedef typename PointVectorT::value_type PointT;

namespace icp
{
namespace point2point
{

template <typename PointVectorT>
void getSourceErrorVector(const PointVectorT& sourcePoints, const std::vector<float>& pointWeights,
		rgbd::eigen::Affine3f const& transform, const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(3 * point_count);

	// point to point.  Need a vector of "T(s)" in: (T(s) - t)^2
	for (unsigned int i = 0; i < point_count; i++) {
		if (correspondence_indices[i] >= 0) {
			PointT transformed_source = transform * sourcePoints[i];
			for (unsigned int j = 0; j < 3; j++) {
				float new_value = pointWeights[i] * transformed_source[j];
				result.push_back(new_value);
			}
		}
	}
}

template <typename PointVectorT>
void getTargetErrorVector(const PointVectorT& targetPoints, const std::vector<float>& pointWeights,
		const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(3 * point_count);

	// point to point.  Need a vector of "t" in: (T(s) - t)^2
	for (unsigned int i = 0; i < point_count; i++) {
		if (correspondence_indices[i] >= 0) {
			PointT target = targetPoints[correspondence_indices[i]];
			for (unsigned int j = 0; j < 3; j++) {
				float new_value = pointWeights[i] * target[j];
				result.push_back(new_value);
			}
		}
	}
}

} // ns
} // ns

#undef POINT_VECTOR_STUFF
