/*
 * depthCodecZlib: (de)compress depth maps using zlib
 *
 * Evan Herbst
 * 3 / 18 / 10
 */

#include <cassert>
#include <zlib.h>
#include "primesensor_tools/depthCodecZlib.h"

namespace zlib
{

/*
 * copy all but the main data buffer and the format flag
 */
void copyHeader(const rgbd_msgs::DepthMap& src, rgbd_msgs::DepthMap& dest)
{
	dest.header = src.header;
	dest.width = src.width;
	dest.height = src.height;
	dest.focal_distance = src.focal_distance;
	dest.no_sample_value = src.no_sample_value;
	dest.shadow_value = src.shadow_value;
}

/*
 * also copy header info
 *
 * pre: in is format_raw
 * post: return value is format_zlib
 */
rgbd_msgs::DepthMapPtr compress(const rgbd_msgs::DepthMap& in)
{
	rgbd_msgs::DepthMapPtr out(new rgbd_msgs::DepthMap);
	compress(in, *out);
	return out;
}
void compress(const rgbd_msgs::DepthMap& in, rgbd_msgs::DepthMap& out)
{
	assert(in.format == rgbd_msgs::DepthMap::format_raw);
	out.format = rgbd_msgs::DepthMap::format_zlib;
	copyHeader(in, out);

	const uLongf rawLen = in.float_data.size() * sizeof(float);
	uLongf destLen = compressBound(rawLen);
	out.binary_data.resize(destLen);
	out.float_data.clear();
	const int ret = compress2((Bytef*)&out.binary_data[0], &destLen, (const Bytef*)&in.float_data[0], rawLen, Z_BEST_SPEED/* TODO make a param */);
	assert(ret == Z_OK);
	out.binary_data.resize(destLen); //deallocate the part of the buffer we aren't using
}

/*
 * also copy header info
 *
 * pre: in is format_zlib
 * post: return value is format_raw
 */
rgbd_msgs::DepthMapPtr decompress(const rgbd_msgs::DepthMap& in)
{
	rgbd_msgs::DepthMapPtr out(new rgbd_msgs::DepthMap);
	decompress(in, *out);
	return out;
}
void decompress(const rgbd_msgs::DepthMap& in, rgbd_msgs::DepthMap& out)
{
	assert(in.format == rgbd_msgs::DepthMap::format_zlib);
	out.format = rgbd_msgs::DepthMap::format_raw;
	copyHeader(in, out);

	uLongf destLen = (out.width * out.height * sizeof(float));
	out.float_data.resize(out.width * out.height);
	const int ret = uncompress((Bytef*)&out.float_data[0], &destLen, (const Bytef*)&in.binary_data[0], in.binary_data.size());
	assert(ret == Z_OK);
}

} //namespace
