#include <stdexcept>

namespace rgbd
{

/*
 * replace NodeHandle::param() for any type for which bool rgbd::ros_param() is defined
 *
 * throw if param server doesn't have it
 */
template <typename T>
void rosparam(ros::NodeHandle& nh, const std::string& name, T& result)
{
	const bool got = ros_param(nh, name, result);
	if(!got) throw std::runtime_error("param '" + name + "' not found");
}

/*
 * usable with any type for which bool rgbd::ros_param() is defined
 *
 * use the default if param server doesn't have it
 */
template <typename T>
void rosparam(ros::NodeHandle& nh, const std::string& name, T& result, const T& defaultVal)
{
	const bool got = ros_param(nh, name, result);
	if(!got) result = defaultVal;
}

} //namespace
