/*
 * sequenceUtils: functions for working with sequences/containers/ranges/whatever you like to call them
 *
 * Evan Herbst
 * 1 / 19 / 12
 */

#include <cstddef>
#include <boost/range/functions.hpp>
#include <boost/range/concepts.hpp>
#include <boost/concept/assert.hpp>

/*
 * copy the selected matrix rows to tgt
 *
 * tgt will be allocated
 */
template <typename SinglePassRangeT>
void copySelectedMtxRows(const rgbd::eigen::MatrixXf& src, const SinglePassRangeT& indices, rgbd::eigen::MatrixXf& tgt)
{
	BOOST_CONCEPT_ASSERT((boost::SinglePassRangeConcept<SinglePassRangeT>));

	tgt.resize(indices.size(), src.cols());
	size_t j = 0;
	for(auto i = boost::begin(indices); i != boost::end(indices); i++) tgt.row(j++) = src.row(*i);
}
